{{- obj|enum_docstring }}
{{ obj.name|upper }} = [
{%- for attr in obj.attrs %}
  {% if not obj|enum_skip(attr.default) -%}
  ("{{ attr.default }}", {{ attr.help }}),
  {%- else -%}
  # ("{{ attr.default }}", {{ attr.help }}),
  {%- endif -%}
{%- endfor %}
]
