# Copyright 2025 Akretion - Raphaël Valyi <raphael.valyi@akretion.com>
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl-3.0.en.html).
# Generated by https://github.com/akretion/xsdata-odoo

import textwrap
from odoo import fields, models
{{ output|default_imports -}}
{%- include "imports.jinja2" -%}
{%- if namespace %}
__NAMESPACE__ = "{{ namespace }}"
{%- endif %}

{{ output }}
